#include "mpi.h"
#include <stdio.h>
#include <math.h>

double f(double a)
{
    return (4.0 / (1.0 + a*a));
}

void main(int argc, char *argv[])
{
    int done = 0, n, myid, numprocs, i;
    double PI25DT = 3.141592653589793238462643;
    double mypi, pi, h, sum, x;
    double startwtime, endwtime;
    int  namelen;
    char processor_name[MPI_MAX_PROCESSOR_NAME];
	
    MPI_Init(&argc,&argv);

	   // Anzahl der Prozesse
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    
	   // meine eigene ID
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    
	   // Prozessename
	MPI_Get_processor_name(processor_name,&namelen);
	
    fprintf(stderr,"Process %d on %s\n",  myid, processor_name);
	fflush(stderr);
	
    n = 0;
    while (!done)
    {
        if (myid == 0)
        {
           printf("Master: Enter the number of intervals: (0 quits) ");
		   fflush(stdout);
	       scanf("%d",&n);
	    
	       startwtime = MPI_Wtime();
        }
           // An alle Client Botschaft senden, wieviele n
			/*
				int MPI_Bcast(	
					void  *smessage,		// Sendepuffer
					int count,
					MPI_Datatype datatype,	 
					int root,				// Sender
					MPI_Comm comm			// Bereichsname
				);
			*/

        MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

        if (n == 0)
            done = 1;  // verabschieden aus Schleife, wenn keine Clients
        else
        {
			puts("Master und Client rechnen");
            h   = 1.0 / (double) n;  // Schrittweite
            sum = 0.0;  // lokale Summe pro Client
			  // Beachte die Incrementierung    i += numprocs
            for (i = myid + 1; i <= n; i += numprocs)
            {
                x = h * ((double)i - 0.5);
                sum += f(x);
            }
            mypi = h * sum;
			mypi = myid+1;
			
			  // Akkumulationsoperation
			/*
				int MPI_Reduce(
					void  *sendbuf,			// sendepuffer			&mypi
					void *recvbuf,									&pi
					int count,										jeder ein double
					MPI_Datatype datatype,							Doubletyp Typ
					MPI_Op op,										Reduce-Op = Summe
					int root,				// Sender				Master ist Sender
					MPI_Comm comm			// Bereichsname			An Alle
				);	
			*/

            MPI_Reduce(&mypi, &pi, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

			
            if (myid == 0)
			{    
				  // nur Master gibt aus
                printf("Master: pi is approximately %.16f, Error is %.16f\n", pi, fabs(pi - PI25DT));
		        endwtime = MPI_Wtime();
		        printf("wall clock time = %f\n", endwtime-startwtime);	       
			}
			else {
              printf("Client: %d mypi is approximately %.16f, Error is %.16f\n", myid, mypi, fabs(mypi - PI25DT));
			}
        }
    }
    MPI_Finalize();
}


